@@ -76,13 +76,16 @@ rsvg_handle_get_pixbuf_sub (RsvgHandle *handle, const char * id)
 
 	g_return_val_if_fail (handle != NULL, NULL);
 
-	if (!handle->finished)
+	if (!handle->priv->finished)
 		return NULL;
 
 	rsvg_handle_get_dimensions (handle, &dimensions);
+	if(!(dimensions.width && dimensions.height))
+		return NULL;
+
 	rowstride = dimensions.width * 4;
 
-	pixels = g_new0(guint8, dimensions.width * dimensions.height * 4);
+	pixels = g_new(guint8, dimensions.width * dimensions.height * 4);
 
 	surface = cairo_image_surface_create_for_data (pixels,
 												   CAIRO_FORMAT_ARGB32,
